/*
 * OHCI HCD (Host Controller Driver) for USB.
 *
 * (C) Copyright 1999 Roman Weissgaerber <weissg@vienna.at>
 * (C) Copyright 2000-2002 David Brownell <dbrownell@users.sourceforge.net>
 * (C) Copyright 2002 Hewlett-Packard Company
 *
 * Bus Glue for NS9750
 *
 * Written by Christopher Hoover <ch@hpl.hp.com>
 * Based on fragments of previous driver by Rusell King et al.
 *
 * Modified for NS9750 from ohci-sa1111.c
 *  by  Jonas Dietsche <jdietsche@fsforth.de>
 *
 * This file is licenced under the GPL.
 */

#include <asm/hardware.h>
#include <asm/mach-types.h>
#include <asm/arch/ns9750_usb.h>
#include <asm/arch/ns9750_bbus.h>
#include <asm/arch/ns9750_bbus_bridge.h>
#include <asm/arch/irqs.h>      /* IRQ_USB_DMA */

#include <asm/io.h>

#define HOST_MODE	0


 extern int usb_disabled(void);

/*-------------------------------------------------------------------------*/

static void ns9750_start_hc( struct platform_device *dev )
{
	printk( KERN_DEBUG "ohci-ns9750.c"
	       ": starting NS9750 OHCI USB Controller\n" );

	/* enable USB Host mode */
	rmwl( NS_BBUS_UTILITY_BASE, | 0x2 );
	writel( NS_BBUS_USB_CFG_CFG_DEVICE | NS_BBUS_USB_CFG_SPEED,
		    NS_BBUS_USB_CFG );
	rmwl( NS_BBUS_UTILITY_BASE, & 0xFFFFFFFD );
	writel( HOST_MODE, NS_USB_MODULE_BASE );

	printk(KERN_DEBUG "ohci-ns9750.c"
		   ": Clock to USB host has been enabled \n");
}

static void ns9750_stop_hc(struct platform_device *dev)
{
	printk(KERN_DEBUG "ohci-ns9750.c"
	       ": stopping NS9750 OHCI USB Controller\n");


}


/*-------------------------------------------------------------------------*/

static irqreturn_t usb_hcd_ns9750_hcim_irq (struct usb_hcd *hcd, struct pt_regs *ptregs)
{
	irqreturn_t ret = ohci_irq( hcd, ptregs );
	/* acknowledge interrupt */
	rmwl( NS_USB_ISTAT, | NS_USB_INT_OHCI_IRQ );
	return ret;
}

/*-------------------------------------------------------------------------*/

void usb_hcd_ns9750_remove (struct usb_hcd *, struct platform_device *);

/* configure so an HC device and id are always provided */
/* always called with process context; sleeping is OK */


/**
 * Allocates basic resources for this USB host controller, and
 * then invokes the start() method for the HCD associated with it
 * through the hotplug entry's driver_data.
 *
 */
int usb_hcd_ns9750_probe (const struct hc_driver *driver,
			  struct platform_device *dev)
{
	int retval;
	struct usb_hcd *hcd = 0;


	hcd = usb_create_hcd(driver, &dev->dev, "ns9750");
	if (hcd == NULL){
		pr_debug ("hcd_alloc failed");
		retval = -ENOMEM;
		goto err1;
	}

	ns9750_start_hc(dev);
	ohci_hcd_init(hcd_to_ohci(hcd));

	hcd->irq =IRQ_USB_HOST;
	hcd->regs = (unsigned int*)( NS_USB_MODULE_BASE + 0x1000 );
	hcd->rsrc_start = NS_USB_MODULE_BASE;
	hcd->self.controller = &dev->dev;

	pr_debug ("%s (NS9750) at 0x%p, irq %d",
	     hcd->driver->descriptiondescription, hcd->regs, hcd->irq);

	hcd->self.bus_name = "ns9750";

	retval = usb_add_hcd(hcd, dev->resource[1].start, SA_INTERRUPT);
	if (retval == 0)
		return retval;

	if ((retval = driver->start (hcd)) < 0)
	{
		usb_hcd_ns9750_remove(hcd, dev);
		return retval;
	}

	return 0;

 err1:
 	kfree(hcd);
	ns9750_stop_hc(dev);
	release_mem_region( NS_USB_MODULE_BASE_PA, 0xB );
	release_mem_region( NS_USB_MODULE_BASE_PA + 0x14, 0xFEC );
	return retval;
}


/* may be called without controller electrically present */
/* may be called with controller, bus, and devices active */

/**
 * usb_hcd_ns9750_remove - shutdown processing for NS9750-based HCDs
 * @dev: USB Host Controller being removed
 * Context: !in_interrupt()
 *
 * Reverses the effect of usb_hcd_ns9750_probe(), first invoking
 * the HCD's stop() method.  It is always called from a thread
 * context, normally "rmmod", "apmd", or something similar.
 *
 */
void usb_hcd_ns9750_remove (struct usb_hcd *hcd, struct platform_device *dev)
{
	pr_debug ("remove: %s, state %x", hcd->self.bus_name, hcd->state);

	if (in_interrupt ())
		BUG ();

	pr_debug ("%s: roothub graceful disconnect", hcd->self.bus_name);
	usb_disconnect (&hcd->self.root_hub);

	hcd->driver->stop (hcd);

	free_irq (hcd->irq, hcd);

	ns9750_stop_hc(dev);

	release_mem_region( NS_USB_MODULE_BASE_PA, 0xB );
	release_mem_region( NS_USB_MODULE_BASE_PA + 0x14, 0xFEC );
}

/*-------------------------------------------------------------------------*/

static int __devinit
ohci_ns9750_start (struct usb_hcd *hcd)
{
	struct ohci_hcd	*ohci = hcd_to_ohci (hcd);
	int		ret;

	ohci_dbg (ohci, "ohci_ns9750_start, ohci:%p", ohci);

	if ((ret = ohci_init(ohci)) < 0)
		return ret;

	if ((ret = ohci_run (ohci)) < 0) {
		err ("can't start %s", hcd->self.bus_name);
		ohci_stop (hcd);
		return ret;
	}

	return 0;
}

/*-------------------------------------------------------------------------*/

static const struct hc_driver ohci_ns9750_hc_driver = {
	.description =		hcd_name,
	.product_desc =	"NS9750 OHCI",
	.hcd_priv_size =	sizeof(struct ohci_hcd),

	/*
	 * generic hardware linkage
	 */
	.irq =		usb_hcd_ns9750_hcim_irq,
	.flags =		HCD_USB11,

	/*
	 * basic lifecycle operations
	 */
	.start =		ohci_ns9750_start,
#ifdef	CONFIG_PM
	/* suspend:		ohci_ns9750_suspend,  -- tbd */
	/* resume:		ohci_ns9750_resume,   -- tbd */
#endif /*CONFIG_PM*/
	.stop =			ohci_stop,


	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue =		ohci_urb_enqueue,
	.urb_dequeue =		ohci_urb_dequeue,
	.endpoint_disable =	ohci_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number =	ohci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data =	ohci_hub_status_data,
	.hub_control =		ohci_hub_control,
};

/*-------------------------------------------------------------------------*/

static int ohci_hcd_ns9750_drv_probe(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct usb_hcd *hcd = NULL;
	int ret;


	if (usb_disabled())
		return -ENODEV;

	ret = usb_hcd_ns9750_probe(&ohci_ns9750_hc_driver, pdev);

	if (ret == 0)
		dev_set_drvdata(dev, hcd);

	return ret;
}

static int ohci_hcd_ns9750_drv_remove(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct usb_hcd *hcd = dev_get_drvdata(dev);

	usb_hcd_ns9750_remove(hcd, pdev);
	dev_set_drvdata(dev, NULL);
	return 0;
}


static struct device_driver ohci_hcd_ns9750_driver = {
	.name		= "ns9750-ohci",
	.bus			= &platform_bus_type,
	.probe		= ohci_hcd_ns9750_drv_probe,
	.remove		= ohci_hcd_ns9750_drv_remove,
	/*.suspend	= ohci_hcd_ns9750_drv_suspend, */
	/*.resume	= ohci_hcd_ns9750_drv_resume, */
};

static int __init ohci_hcd_ns9750_init (void)
{
	pr_debug (DRIVER_INFO " (NS9750)");
	pr_debug ("block sizes: ed %d td %d\n",
		sizeof (struct ed), sizeof (struct td));


        if( !request_mem_region( NS_USB_MODULE_BASE_PA, 0xB, hcd_name) )
        {
                printk( "request_mem_region failed\n" );
                return -EBUSY;
        }

        if( !request_mem_region( NS_USB_MODULE_BASE_PA + 0x14, 0xFEC, hcd_name) )
        {
                printk( "request_mem_region failed\n" );
                return -EBUSY;
        }


	return driver_register(&ohci_hcd_ns9750_driver);
}

static void __exit ohci_hcd_ns9750_cleanup (void)
{
	driver_unregister(&ohci_hcd_ns9750_driver);
}

module_init (ohci_hcd_ns9750_init);
module_exit (ohci_hcd_ns9750_cleanup);
